# CollectX API


The CollectX API offers a fast and convenient way to process event/counter data in the CollectX style.

Several scenarios are available:
1. write data as binary files.
2. transfer data to clx (collector) process using IPC (Inter Process Comunication)
3. export data through Fluent Bit
4. create Prometheus endpoint and keep the most recent data to be scraped by Prometheus.

API stores data in the internal buffer and flushes it to destinations once it full, on user demand, or at the end of the run.

The data consists of 2 parts.  A schema containing metadata, and the actual binary data. For the file write and IPC cases both schema and binary data will be transfered. For Fluent-Bit and Prometheus exporters data will be converted to the formats required by destinations.

## Getting started
Available types and functions are defined in `Collectx/src/api/clx_api.h` (installed under `/opt/mellanox/collectx/include/api/clx_api.h`). API is built as shared object library `libclx_api.so`.

The easiest way to start is to copy one of the examples from "Collectx/src/api/api_tester" (installed under "/opt/mellanox/collectx/examples"):
* clx_api_counter_example.c
* clx_api_event_example.c

Both examples documented in-place.

libclx_api.so is installed under opt/mellanox/collectx/lib .

To build examples:
```
$ cd /opt/mellanox/collectx/examples/api
# build counters example
$ sudo gcc -std=gnu99 -I/opt/mellanox/collectx/include/ -L/opt/mellanox/collectx/lib/ -o clx_api_counter_example  clx_api_counter_example.c -lclx_api -lz

#build events example
$ sudo gcc -std=gnu99 -I/opt/mellanox/collectx/include/ -L/opt/mellanox/collectx/lib/ -o clx_api_event_example  clx_api_event_example.c -lclx_api -lz
```

To run built examples:
```
$ sudo LD_LIBRARY_PATH=/opt/mellanox/collectx/lib ./clx_api_counter_example
$ sudo LD_LIBRARY_PATH=/opt/mellanox/collectx/lib ./clx_api_event_example
```

The following subsections contain detailed explanation of clx_api.

### Initialization
To start using clx_api, user need to
1. set API parameters `p` of type `clx_api_params_t *`
2. create `provider` of type `clx_api_provider_t *`
3. create API context `void* clx_api_create_context(clx_api_params_t* p, clx_api_provider_t* provider)`

API parameters contain the configuration of data destination and internal buffer size `buffer_size`.
Provider's `version`, `name`, `description` and `.initialize`-callback should be set.
`.initialize`-calback should use one of 2 options:
1. create event types using `clx_api_add_event_type`,
2. create counters using `clx_api_add_counter`.
using both options at the same time is not allowed.

Created events/counters will be converted to schema.
Do not add events/counters after this step.

### Collecting the data from several sources

Data can be collected from different sources. To set the following fields of API parameters:
* `source_id`                      - {source} in data_path_template. Set to node, guid, etc.
* `source_tag`                     - {tag} in data_path_template. Use empty-string tag for counters and non-empty for events.

In the scenario when the similar data should be collected from several sources running as a single process, context should be cloned from original source.
Call `clone = clx_api_clone_context(orig_ctx, new_params)` to clone the context.

## Destinations
### File write

The data consists of 2 parts.  A schema containing metadata, and the actual binary data.
* `data_root` (absolute/relative)   - path to directory in which to save bin data files.
* `schema_path` (absolute/relative) - path to directory in which to save schema files.

* `data_path_template` - defines the naming convention to be used when writing files.

When the file write enabled from parameters, data will be saved to binary file according to naming convention. The following tree shows the default naming convention `CLX_DATA_PATH_TEMPLATE="{{year}}/{{month}}{{day}}/{{source}}/{{tag}}{{id}}.bin"`:
~~~
data_root
├── {year}
│   └── {mmdd}
│       ├── {source_id}
│       │   └── {source_tag}{timestamp}.bin
│       └── {another_source_id}
│           └── {another_source_tag}{timestamp}.bin
└── schema
    └── schema_{MD5_digest}.json
~~~

Once one of 2 limits (`max_file_size` or `max_file_age`) is reached, the current binary file will be closed and the next file will be generated.

To set up file write specify the following parameters:
* `max_file_size` (bytes)        - max size for bin file in
* `max_file_age`  (microseconds) - if file is older than `max_file_age`, a new file will be created.

## IPC transport
When IPC transport is enabled, the data will be send from clx_api process to clx process via shared memory.

To enable IPC configure API parameters on the initialization step:
* `ipc_enabled                = 1;`
* `ipc_sockets_dir            ="/tmp/ipc_sockets";  // default. should contain clx ipc socket`

Optionaly set timeouts for reattaching and IPC sockets:
* `ipc_max_reattach_time_msec = 10000` // default is 10 seconds for all reatach tries
* `ipc_max_reattach_tries     = 10`    // default is 10 tries during the max_reattach_time
* `ipc_socket_timeout_msec    = 3000`  // default is 3 sec. Timeout is used for send/reply procedure.
See the Collectx/src/ipc_provider/README.md for more details of IPC.

Note that IPC transport exploits system folders. Run clx_api application with sudo to be able to use IPC.

Usually clx stays online for a long periods, but it may restarted for some reason or it may start after the clx_api based application.
To check the status of IPC for current context use `clx_api_get_ipc_status(api_ctx)`.
To make sure that IPC is connected use `clx_api_ipc_connect_if_needed(api_ctx)`.
Note that in the future function `clx_api_ipc_connect_if_needed(api_ctx)` will be part of implementation, not API.

## Fluent Bit exporter

Make sure that clx was build with fluent-bit, that requires cmake3.

Create folder fluent_bit_configs with exp-files for each destination. Use Collectx/deploy/ufm/etc/fluent_bit_configs as a reference. All exp-files are documented in-place.

Set environment variables before running:
1. `FLUENT_BIT_EXPORT_ENABLE=1`
2. `FLUENT_BIT_CONFIG_DIR=/path/to/fluent_bit_configs`
3. `set LD_LIBRARY_PATH=/opt/mellanox/collectx/lib  if clx installed with fluent-bit`
   or
   set path to `FLUENT_BIT_EXPORT_API_SO=/path/to/libraw_msgpack_api.so`

## Prometheus exporter
Prometheus exporter sets up endpoint (HTTP server) which will keep the most recent counters / events data.
Prometheus server will be able to scrape the data from the endpoint while clx_api application stays online.

See the generic example of prometheus records:
~~~
counter_name_1{label_1="label_1_val", label_2="label_2_val", label_3="label_3_val", label_4="label_4_val"} counter_value_1 timestamp_1
counter_name_2{label_1="label_1_val", label_2="label_2_val", label_3="label_3_val", label_4="label_4_val"} counter_value_2 timestamp_2
...
~~~
Labels are customizable metadata which can be set from data file. Counters / events names could be filtered by token based name-match according to countersets / fieldsets.

Set the following environment variables before running.

| Variable | Description |
| -------- | ----------- |
| `PROMETHEUS_ENDPOINT=http://0.0.0.0:9101` | Set the enpoint host and port to enable export. |
| `PROMETHEUS_INDEXES=port_num` | Set indexes as a comma-separated list to keep data for every index field. In this example most recent data will be kept for every record with unique `port_num`. If not set, only one data per source will be kept as the most recent. |
| `CLX_METADATA_FILE=/path/to/labels.txt` | Set path to a file with prometheus a custom labels. Use labels to store information about data source and indexes. If not set, the default labels will be used. |
|`PROMETHEUS_CSET_DIR=/path/to/prometheus_cset` | Set the folder which contains cset and fset files. If set, Prometheus will scrape only filtered data according to countersets/fieldsets. |
<!-- TBD: this is only for UFM telemetry docker? if not, uncomment -->
<!-- |CLX_METADATA_COMMAND=gen_metadata --fabric compute --file /tmp/ibd/ibdiagnet2.ibnetdiscover --output labels.txt | Set the user-implemented command that will generate labels on the fly | -->


Prometheus labels could be iether obtained from file

Tip: to scrape the data without Prometheus server use:
~~~
curl -s http://0.0.0.0:9101/metrics
~~~
or
~~~
curl -s http://0.0.0.0:9101/{cset_name}
~~~

# API functions
The following table shows top level description of the API functions.
| Signature | Description  |
| ----------| ------------ |
| INITIALIZE / DESTROY FUNCTIONS ||
| `void* clx_api_create_context(clx_api_params_t* p, clx_api_provider_t* provider)`| Creates API context based on parameters and provider. |
| `void* clx_api_clone_context(void* vctx, clx_api_params_t* p)`                   | Clones initial context vctx according to new parameters. Cloned contexts use the same provider, e.i. it will share the same schema. Use clones when the same data should be collected from different sources. Each clone will write the data to its own file, while all clones will share the same exporters. |
| `bool clx_api_destroy_context(void * vctx)` | Destroys clx_api context. Should be applied to all clones and original context. |
| | |
| CONFIGURATION ACCESS FUNCTIONS | |
| `clx_api_params_t* clx_api_get_params(void* vctx)`                               | Get the parameters of context / clone `vctx` |
| `bool clx_api_is_file_write_enabled(void* vctx)`                                 | Check if file write is enabled |
| `uint64_t clx_api_get_file_timestamp(void* vctx)`                                | Get timestamp of the context/clone binary file|
| `char* clx_api_get_data_root(void* vctx)`                                        | Get current data root of the context / clone `vctx`|
| | |
| DATA FUNCTIONS ||
| `void clx_api_force_write(void* vctx)` | Flush data to all sources immediately. |
| EVENTS DATA FUNCTIONS ||
| `bool clx_api_add_event_type(void *ctx, char *event_name, clx_api_event_field_info_t* fields, int num_fields, uint8_t* index)` | To create event type of name `event_name` with number of fields `num_fields` and get `index` of the event in the schema. Create all events only in the  provider `.initialize`-callback. |
| `bool clx_api_write_event(void* ctx, void* event_data, clx_api_type_index_t ti, int num_events)` | Write `num_events` events of index `ti` to clx_api internal buffer. |
| COUNTERS DATA FUNCTIONS ||
| `bool clx_api_add_counter(void* ctx, clx_api_counter_info_t* counter, char* group_name, uint32_t* counter_num)`| To create new counter and add it to schema group. Create all counters only in the  provider `.initialize`-callback. |
| `void* clx_api_get_counters_buffer(void* ctx, uint64_t timestamp, uint32_t *data_size)` | Get counters buffer to write counters manually. Counters buffer will be created as a free memory segment of clx_api internal buffer. If there is not enough space, it will flush previously written data. Returns `NULL` on error. |
| | |
| UTILS||
| `uint64_t clx_api_get_timestamp(void)` | get the current time in CollectX format. |
| `void clx_api_set_first_timestamp(void* vctx, uint64_t timestamp)` | Use to initialize first binary file start timestamp to non-default value if needed. If the app that uses clx_api will restart, it will write to the new file. To continue to write to the previous file reset timestamp according to the timestamp in the previous file. |
| `clx_api_get_ipc_status(void* vctx)` | Return status of IPC transport for api context. Returns -1 if IPC is disabled from config, 0 if IPC is connected, 1 if latest IPC message is not delivered due to client disconnection. |
| `void clx_api_ipc_connect_if_needed(void* ctx)` | Use to make sure that IPC connected. If clx process restarted for some reason it will reestablish the connection. Later this function will be part of implementation, not API. |
